from __future__ import annotations

from typing import List, cast

import grimp
from grimp import ImportGraph
from typing_extensions import TypedDict

from importlinter.application import contract_utils, output
from importlinter.application.contract_utils import AlertLevel
from importlinter.domain import fields
from importlinter.domain.contract import Contract, ContractCheck
from importlinter.domain.imports import Module

from ._common import DetailedChain, Link, build_detailed_chain_from_route, render_chain_data


class _SubpackageChainData(TypedDict):
    upstream_module: str
    downstream_module: str
    chains: List[DetailedChain]


class IndependenceContract(Contract):
    """
    Independence contracts check that a set of modules do not depend on each other.

    They do this by checking that there are no imports in any direction between the modules,
    even indirectly.

    Configuration options:

        - modules:        A list of Modules that should be independent of each other.
        - ignore_imports: A set of ImportExpressions. These imports will be ignored: if the import
                          would cause a contract to be broken, adding it to the set will cause
                          the contract be kept instead. (Optional.)
        - unmatched_ignore_imports_alerting: Decides how to report when the expression in the
                          `ignore_imports` set is not found in the graph. Valid values are
                          "none", "warn", "error". Default value is "error".
    """

    type_name = "independence"

    modules = fields.ListField(subfield=fields.ModuleField())
    ignore_imports = fields.SetField(subfield=fields.ImportExpressionField(), required=False)
    unmatched_ignore_imports_alerting = fields.EnumField(AlertLevel, default=AlertLevel.ERROR)

    def check(self, graph: ImportGraph, verbose: bool) -> ContractCheck:
        warnings = contract_utils.remove_ignored_imports(
            graph=graph,
            ignore_imports=self.ignore_imports,  # type: ignore
            unmatched_alerting=self.unmatched_ignore_imports_alerting,  # type: ignore
        )

        self._check_all_modules_exist_in_graph(graph)

        dependencies = graph.find_illegal_dependencies_for_layers(
            # A single layer consisting of siblings.
            layers=({module.name for module in self.modules},),  # type: ignore
        )
        invalid_chains = self._build_invalid_chains(dependencies, graph)

        return ContractCheck(
            kept=not dependencies,
            warnings=warnings,
            metadata={"invalid_chains": invalid_chains},
        )

    def render_broken_contract(self, check: "ContractCheck") -> None:
        for chains_data in cast(List[_SubpackageChainData], check.metadata["invalid_chains"]):
            downstream, upstream = (
                chains_data["downstream_module"],
                chains_data["upstream_module"],
            )
            output.print(f"{downstream} is not allowed to import {upstream}:")
            output.new_line()

            for chain_data in chains_data["chains"]:
                render_chain_data(chain_data)
                output.new_line()

            output.new_line()

    def _check_all_modules_exist_in_graph(self, graph: ImportGraph) -> None:
        for module in self.modules:  # type: ignore
            if module.name not in graph.modules:
                raise ValueError(f"Module '{module.name}' does not exist.")

    def _build_invalid_chains(
        self, dependencies: set[grimp.PackageDependency], graph: grimp.ImportGraph
    ) -> list[_SubpackageChainData]:
        return [
            {
                "upstream_module": dependency.imported,
                "downstream_module": dependency.importer,
                "chains": [build_detailed_chain_from_route(c, graph) for c in dependency.routes],
            }
            for dependency in dependencies
        ]

    def _build_subpackage_chain_data(
        self, upstream_module: Module, downstream_module: Module, graph: ImportGraph
    ) -> _SubpackageChainData:
        """
        Return any import chains from the upstream to downstream module.
        """
        subpackage_chain_data: _SubpackageChainData = {
            "upstream_module": upstream_module.name,
            "downstream_module": downstream_module.name,
            "chains": [],
        }
        assert isinstance(subpackage_chain_data["chains"], list)  # For type checker.
        chains = graph.find_shortest_chains(
            importer=downstream_module.name, imported=upstream_module.name
        )
        if chains:
            for chain in chains:
                chain_data: List[Link] = []
                for importer, imported in [
                    (chain[i], chain[i + 1]) for i in range(len(chain) - 1)
                ]:
                    import_details = graph.get_import_details(importer=importer, imported=imported)
                    line_numbers = tuple(j["line_number"] for j in import_details)
                    chain_data.append(
                        {
                            "importer": importer,
                            "imported": imported,
                            "line_numbers": line_numbers,
                        }
                    )
                detailed_chain: DetailedChain = {
                    "chain": chain_data,
                    "extra_firsts": [],
                    "extra_lasts": [],
                }
                subpackage_chain_data["chains"].append(detailed_chain)

        return subpackage_chain_data



from typing import Iterable, List

from importlinter.application.app_config import settings


def find_any(*filenames: Iterable[str]) -> List[str]:
    """
    Return a list of names of any potential files that contain config.

    Args:
        *filenames: list of filenames, e.g. ('setup.cfg', '.importlinter').

    Returns:
        List of absolute filenames that could be found.
    """
    found_files: List[str] = []

    filesystem = settings.FILE_SYSTEM
    current_working_directory = filesystem.getcwd()

    for filename in filenames:
        candidate_filename = filesystem.join(current_working_directory, filename)

        if filesystem.exists(candidate_filename):
            found_files.append(candidate_filename)

    return found_files



from .adapters.building import GraphBuilder
from .adapters.filesystem import FileSystem
from .adapters.printing import ClickPrinter
from .adapters.timing import SystemClockTimer
from .adapters.user_options import IniFileUserOptionReader, TomlFileUserOptionReader
from .application.app_config import settings


def configure():
    settings.configure(
        USER_OPTION_READERS={
            "ini": IniFileUserOptionReader(),
            "toml": TomlFileUserOptionReader(),
        },
        GRAPH_BUILDER=GraphBuilder(),
        PRINTER=ClickPrinter(),
        FILE_SYSTEM=FileSystem(),
        TIMER=SystemClockTimer(),
        DEFAULT_CACHE_DIR=".import_linter_cache",
    )



__version__ = "2.0"

from .application import output  # noqa
from .domain import fields  # noqa
from .domain.contract import Contract, ContractCheck  # noqa

__all__ = ["output", "fields", "Contract", "ContractCheck"]



from typing import Any, Dict, List


class InvalidUserOptions(Exception):
    pass


class UserOptions:
    """
    Configuration supplied by the end user.

    Arguments:
        - session_options:   General options relating to the running of the linter.
        - contracts_options: List of the options that will be used to build the contracts.
    """

    def __init__(
        self, session_options: Dict[str, Any], contracts_options: List[Dict[str, Any]]
    ) -> None:
        self.session_options = session_options
        self.contracts_options = contracts_options

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, UserOptions):
            return False
        return (self.session_options == other.session_options) and (
            self.contracts_options == other.contracts_options
        )



import importlib
from copy import copy, deepcopy
from typing import Any, Dict, List, Optional, Tuple, Type, Union

from grimp import ImportGraph

from ..application import rendering
from ..domain.contract import Contract, InvalidContractOptions, registry
from . import output
from .app_config import settings
from .ports.reporting import Report
from .rendering import render_exception, render_report
from .sentinels import NotSupplied
from .user_options import UserOptions

# Public functions
# ----------------

SUCCESS = True
FAILURE = False


def lint_imports(
    config_filename: Optional[str] = None,
    limit_to_contracts: Tuple[str, ...] = (),
    cache_dir: Union[str, None, Type[NotSupplied]] = NotSupplied,
    is_debug_mode: bool = False,
    show_timings: bool = False,
    verbose: bool = False,
) -> bool:
    """
    Analyse whether a Python package follows a set of contracts, and report on the results.

    This function attempts to handle and report all exceptions, too.

    Args:
        config_filename:    the filename to use to parse user options.
        limit_to_contracts: if supplied, only lint the contracts with the supplied ids.
        cache_dir:          the directory to use for caching. Pass None to disable caching.
        is_debug_mode:      whether debugging should be turned on. In debug mode, exceptions are
                            not swallowed at the top level, so the stack trace can be seen.
        show_timings:       whether to show the times taken to build the graph and to check
                            each contract.
        verbose:            if True, noisily output progress as it goes along.

    Returns:
        True if the linting passed, False if it didn't.
    """
    output.print_heading("Import Linter", output.HEADING_LEVEL_ONE)
    output.verbose_print(verbose, "Verbose mode.")
    try:
        user_options = read_user_options(config_filename=config_filename)
        _register_contract_types(user_options)
        report = create_report(user_options, limit_to_contracts, cache_dir, show_timings, verbose)
    except Exception as e:
        if is_debug_mode:
            raise e
        render_exception(e)
        return FAILURE

    render_report(report)

    if report.contains_failures:
        return FAILURE
    else:
        return SUCCESS


def read_user_options(config_filename: Optional[str] = None) -> UserOptions:
    """
    Return the UserOptions object from the supplied config file.

    If no filename is supplied, look in the default location
    (see importlinter.cli.lint_imports).

    Raises:
        FileNotFoundError if no configuration file could be found.
    """
    readers = settings.USER_OPTION_READERS.values()
    if config_filename:
        if config_filename.endswith(".toml"):
            readers = [settings.USER_OPTION_READERS["toml"]]
        else:
            readers = [settings.USER_OPTION_READERS["ini"]]

    for reader in readers:
        options = reader.read_options(config_filename=config_filename)
        if options:
            normalized_options = _normalize_user_options(options)
            return normalized_options
    raise FileNotFoundError("Could not read any configuration.")


def create_report(
    user_options: UserOptions,
    limit_to_contracts: Tuple[str, ...] = tuple(),
    cache_dir: Union[str, None, Type[NotSupplied]] = NotSupplied,
    show_timings: bool = False,
    verbose: bool = False,
) -> Report:
    """
    Analyse whether a Python package follows a set of contracts, returning a report on the results.

    Raises:
        InvalidUserOptions: if the report could not be run due to invalid user configuration,
                            such as a module that could not be imported.
    """
    include_external_packages = _get_include_external_packages(user_options)
    exclude_type_checking_imports = _get_exclude_type_checking_imports(user_options)

    with settings.TIMER as timer:
        graph = _build_graph(
            root_package_names=user_options.session_options["root_packages"],
            cache_dir=cache_dir,
            include_external_packages=include_external_packages,
            exclude_type_checking_imports=exclude_type_checking_imports,
            verbose=verbose,
        )
    graph_building_duration = timer.duration_in_s
    output.verbose_print(verbose, f"Built graph in {graph_building_duration}s.")

    return _build_report(
        graph=graph,
        graph_building_duration=graph_building_duration,
        user_options=user_options,
        limit_to_contracts=limit_to_contracts,
        show_timings=show_timings,
        verbose=verbose,
    )


# Private functions
# -----------------


def _normalize_user_options(user_options: UserOptions) -> UserOptions:
    normalized_options = copy(user_options)
    if "root_packages" not in normalized_options.session_options:
        normalized_options.session_options["root_packages"] = [
            normalized_options.session_options["root_package"]
        ]
    if "root_package" in normalized_options.session_options:
        del normalized_options.session_options["root_package"]
    return normalized_options


def _build_graph(
    root_package_names: List[str],
    include_external_packages: Optional[bool],
    exclude_type_checking_imports: bool,
    verbose: bool,
    cache_dir: Union[str, None, Type[NotSupplied]] = NotSupplied,
) -> ImportGraph:
    if cache_dir == NotSupplied:
        cache_dir = settings.DEFAULT_CACHE_DIR

    if cache_dir:
        output.verbose_print(verbose, f"Building import graph (cache directory is {cache_dir})...")
    else:
        output.verbose_print(verbose, "Building import graph (with caching disabled)...")

    return settings.GRAPH_BUILDER.build(
        root_package_names=root_package_names,
        include_external_packages=include_external_packages,
        exclude_type_checking_imports=exclude_type_checking_imports,
        cache_dir=cache_dir,
    )


def _build_report(
    graph: ImportGraph,
    graph_building_duration: int,
    user_options: UserOptions,
    limit_to_contracts: Tuple[str, ...],
    show_timings: bool,
    verbose: bool,
) -> Report:
    report = Report(
        graph=graph, show_timings=show_timings, graph_building_duration=graph_building_duration
    )
    contracts_options = _filter_contract_options(
        user_options.contracts_options, limit_to_contracts
    )
    for contract_options in contracts_options:
        contract_class = registry.get_contract_class(contract_options["type"])
        try:
            contract = contract_class(
                name=contract_options["name"],
                session_options=user_options.session_options,
                contract_options=contract_options,
            )
        except InvalidContractOptions as e:
            report.add_invalid_contract_options(contract_options["name"], e)
            return report

        output.verbose_print(verbose, f"Checking {contract.name}...")
        with settings.TIMER as timer:
            # Make a copy so that contracts can mutate the graph without affecting
            # other contract checks.
            copy_of_graph = deepcopy(graph)
            check = contract.check(copy_of_graph, verbose=verbose)
        report.add_contract_check(contract, check, duration=timer.duration_in_s)
        if verbose:
            rendering.render_contract_result_line(contract, check, duration=timer.duration_in_s)

    output.verbose_print(verbose, newline=True)
    return report


def _filter_contract_options(
    contracts_options: List[Dict[str, Any]], limit_to_contracts: Tuple[str, ...]
) -> List[Dict[str, Any]]:
    if limit_to_contracts:
        # Validate the supplied contract ids.
        registered_contract_ids = {option["id"] for option in contracts_options}
        missing_contract_ids = set(limit_to_contracts) - registered_contract_ids
        if missing_contract_ids:
            if len(missing_contract_ids) == 1:
                raise ValueError(
                    f"Could not find contract '{missing_contract_ids.pop()}'.\n\n"
                    "You asked to limit the check to that contract, but nothing exists "
                    "with that id."
                )
            else:
                raise ValueError(
                    "Could not find the following contract ids: "
                    f"{', '.join(sorted(missing_contract_ids))}.\n\n"
                    "You asked to limit the check to those contracts, but there are no "
                    "contracts with those ids."
                )
        else:
            return [o for o in contracts_options if o["id"] in limit_to_contracts]
    else:
        return contracts_options


def _register_contract_types(user_options: UserOptions) -> None:
    contract_types = _get_built_in_contract_types() + _get_plugin_contract_types(user_options)
    for name, contract_class in contract_types:
        registry.register(contract_class, name)


def _get_built_in_contract_types() -> List[Tuple[str, Type[Contract]]]:
    return list(
        map(
            _parse_contract_type_string,
            [
                "forbidden: importlinter.contracts.forbidden.ForbiddenContract",
                "layers: importlinter.contracts.layers.LayersContract",
                "independence: importlinter.contracts.independence.IndependenceContract",
            ],
        )
    )


def _get_plugin_contract_types(user_options: UserOptions) -> List[Tuple[str, Type[Contract]]]:
    contract_types = []
    if "contract_types" in user_options.session_options:
        for contract_type_string in user_options.session_options["contract_types"]:
            contract_types.append(_parse_contract_type_string(contract_type_string))
    return contract_types


def _parse_contract_type_string(string) -> Tuple[str, Type[Contract]]:
    components = string.split(": ")
    assert len(components) == 2
    name, contract_class_string = components
    contract_class = _string_to_class(contract_class_string)
    if not issubclass(contract_class, Contract):
        raise TypeError(f"{contract_class} is not a subclass of Contract.")
    return name, contract_class


def _string_to_class(string: str) -> Type:
    """
    Parse a string into a Python class.

    Args:
        string: a fully qualified string of a class, e.g. 'mypackage.foo.MyClass'.

    Returns:
        The class.
    """
    components = string.split(".")
    class_name = components[-1]
    module_name = ".".join(components[:-1])
    module = importlib.import_module(module_name)
    cls = getattr(module, class_name)
    assert isinstance(cls, type)
    return cls


def _get_include_external_packages(user_options: UserOptions) -> Optional[bool]:
    """
    Get a boolean (or None) for the include_external_packages option in user_options.
    """
    try:
        include_external_packages_str = user_options.session_options["include_external_packages"]
    except KeyError:
        return None
    # Cast the string to a boolean.
    return include_external_packages_str in ("True", "true")


def _get_exclude_type_checking_imports(user_options: UserOptions) -> bool:
    """
    Get a boolean for the exclude_type_checking_imports option in user_options.
    """
    try:
        exclude_type_checking_imports_str = user_options.session_options[
            "exclude_type_checking_imports"
        ]
    except KeyError:
        return False
    # Cast the string to a boolean.
    return exclude_type_checking_imports_str in ("True", "true")


def _get_show_timings(user_options: UserOptions) -> bool:
    """
    Get a boolean (or None) for the show_timings option in user_options.
    """
    try:
        show_timings_str = user_options.session_options["show_timings"]
    except KeyError:
        return False
    # Cast the string to a boolean.
    return show_timings_str in ("True", "true")



import abc
from typing import Optional


class Printer(abc.ABC):
    @abc.abstractmethod
    def print(
        self, text: str = "", bold: bool = False, color: Optional[str] = None, newline: bool = True
    ) -> None:
        """
        Prints a line.
        """
        raise NotImplementedError



"""
Helper functions used by more than one contract type.

Code in here should not be relied upon as a public API; if you're
relying on it for a custom contract type, be aware things may change
without warning.
"""
from __future__ import annotations

import itertools
from typing import List, Optional, Sequence, Tuple, Union

import grimp
from grimp import ImportGraph
from typing_extensions import TypedDict

from importlinter.application import output
from importlinter.domain.imports import Module


class Link(TypedDict):
    importer: str
    imported: str
    # If the graph has been built manually, we may not know the line number.
    line_numbers: Tuple[Optional[int], ...]


Chain = List[Link]


class DetailedChain(TypedDict):
    chain: Chain
    extra_firsts: List[Link]
    extra_lasts: List[Link]


def render_chain_data(chain_data: DetailedChain) -> None:
    main_chain = chain_data["chain"]
    _render_direct_import(main_chain[0], extra_firsts=chain_data["extra_firsts"], first_line=True)

    for direct_import in main_chain[1:-1]:
        _render_direct_import(direct_import)

    if len(main_chain) > 1:
        _render_direct_import(main_chain[-1], extra_lasts=chain_data["extra_lasts"])


def find_segments(
    graph: ImportGraph, reference_graph: ImportGraph, importer: Module, imported: Module
) -> List[Chain]:
    """
    Return list of headless and tailless chains.

    Two graphs are passed in: the first is mutated, the second is used purely as a reference to
    look up import details which are otherwise removed during mutation.
    """
    segments = []
    for chain in _pop_shortest_chains(graph, importer=importer.name, imported=imported.name):
        if len(chain) == 2:
            raise ValueError("Direct chain found - these should have been removed.")
        segment: List[Link] = []
        for importer_in_chain, imported_in_chain in [
            (chain[i], chain[i + 1]) for i in range(len(chain) - 1)
        ]:
            import_details = reference_graph.get_import_details(
                importer=importer_in_chain, imported=imported_in_chain
            )
            line_numbers = tuple(sorted(set(j["line_number"] for j in import_details)))
            segment.append(
                {
                    "importer": importer_in_chain,
                    "imported": imported_in_chain,
                    "line_numbers": line_numbers,
                }
            )
        segments.append(segment)
    return segments


def segments_to_collapsed_chains(
    graph: ImportGraph, segments: List[Chain], importer: Module, imported: Module
) -> List[DetailedChain]:
    collapsed_chains: List[DetailedChain] = []
    for segment in segments:
        head_imports: List[Link] = []
        imported_module = segment[0]["imported"]
        candidate_modules = sorted(graph.find_modules_that_directly_import(imported_module))
        for module in [
            m
            for m in candidate_modules
            if Module(m) == importer or Module(m).is_descendant_of(importer)
        ]:
            import_details_list = graph.get_import_details(
                importer=module, imported=imported_module
            )
            line_numbers = tuple(sorted(set(j["line_number"] for j in import_details_list)))
            head_imports.append(
                {"importer": module, "imported": imported_module, "line_numbers": line_numbers}
            )

        tail_imports: List[Link] = []
        importer_module = segment[-1]["importer"]
        candidate_modules = sorted(graph.find_modules_directly_imported_by(importer_module))
        for module in [
            m
            for m in candidate_modules
            if Module(m) == imported or Module(m).is_descendant_of(imported)
        ]:
            import_details_list = graph.get_import_details(
                importer=importer_module, imported=module
            )
            line_numbers = tuple(sorted(set(j["line_number"] for j in import_details_list)))
            tail_imports.append(
                {"importer": importer_module, "imported": module, "line_numbers": line_numbers}
            )

        collapsed_chains.append(
            {
                "chain": [head_imports[0]] + segment[1:-1] + [tail_imports[0]],
                "extra_firsts": head_imports[1:],
                "extra_lasts": tail_imports[1:],
            }
        )

    return collapsed_chains


def _pop_shortest_chains(graph: ImportGraph, importer: str, imported: str):
    chain: Union[Optional[Tuple[str, ...]], bool] = True
    while chain:
        chain = graph.find_shortest_chain(importer, imported)
        if chain:
            # Remove chain of imports from graph.
            for index in range(len(chain) - 1):
                graph.remove_import(importer=chain[index], imported=chain[index + 1])
            yield chain


def format_line_numbers(line_numbers: Sequence[Optional[int]]) -> str:
    """
    Return a human-readable string of the supplied line numbers.

    Unknown line numbers should be provided as a None value in the sequence. E.g.
    (None,) will be returned as "l.?".
    """
    return ", ".join(
        "l.?" if line_number is None else f"l.{line_number}" for line_number in line_numbers
    )


def _render_direct_import(
    direct_import,
    first_line: bool = False,
    extra_firsts: Optional[List] = None,
    extra_lasts: Optional[List] = None,
) -> None:
    import_strings = []
    if extra_firsts:
        for position, source in enumerate([direct_import] + extra_firsts[:-1]):
            prefix = "& " if position > 0 else ""
            importer = source["importer"]
            line_numbers = format_line_numbers(source["line_numbers"])
            import_strings.append(f"{prefix}{importer} ({line_numbers})")
        importer, imported = extra_firsts[-1]["importer"], extra_firsts[-1]["imported"]
        line_numbers = format_line_numbers(extra_firsts[-1]["line_numbers"])
        import_strings.append(f"& {importer} -> {imported} ({line_numbers})")
    else:
        importer, imported = direct_import["importer"], direct_import["imported"]
        line_numbers = format_line_numbers(direct_import["line_numbers"])
        import_strings.append(f"{importer} -> {imported} ({line_numbers})")

    if extra_lasts:
        indent_string = (len(direct_import["importer"]) + 4) * " "
        for destination in extra_lasts:
            imported = destination["imported"]
            line_numbers = format_line_numbers(destination["line_numbers"])
            import_strings.append(f"{indent_string}& {imported} ({line_numbers})")

    for position, import_string in enumerate(import_strings):
        if first_line and position == 0:
            output.print_error(f"- {import_string}", bold=False)
        else:
            output.print_error(f"  {import_string}", bold=False)


def build_detailed_chain_from_route(route: grimp.Route, graph: grimp.ImportGraph) -> DetailedChain:
    ordered_heads = sorted(route.heads)
    extra_firsts: list[Link] = [
        {
            "importer": head,
            "imported": route.middle[0],
            "line_numbers": get_line_numbers(importer=head, imported=route.middle[0], graph=graph),
        }
        for head in ordered_heads[1:]
    ]
    ordered_tails = sorted(route.tails)
    extra_lasts: list[Link] = [
        {
            "imported": tail,
            "importer": route.middle[-1],
            "line_numbers": get_line_numbers(
                imported=tail, importer=route.middle[-1], graph=graph
            ),
        }
        for tail in ordered_tails[1:]
    ]
    chain_as_strings = [ordered_heads[0], *route.middle, ordered_tails[0]]
    chain_as_links: Chain = [
        {
            "importer": importer,
            "imported": imported,
            "line_numbers": get_line_numbers(importer=importer, imported=imported, graph=graph),
        }
        for importer, imported in pairwise(chain_as_strings)
    ]
    return {
        "chain": chain_as_links,
        "extra_firsts": extra_firsts,
        "extra_lasts": extra_lasts,
    }


def get_line_numbers(
    importer: str, imported: str, graph: grimp.ImportGraph
) -> tuple[int | None, ...]:
    details = graph.get_import_details(importer=importer, imported=imported)
    line_numbers = tuple(i["line_number"] for i in details) if details else (None,)
    return line_numbers


def pairwise(iterable):
    """
    Return successive overlapping pairs taken from the input iterable.
    pairwise('ABCDEFG') --> AB BC CD DE EF FG

    TODO: Replace with itertools.pairwise once on Python 3.10.
    """
    a, b = itertools.tee(iterable)
    next(b, None)
    return zip(a, b)



from __future__ import annotations

import abc
from types import TracebackType


class Timer(abc.ABC):
    """
    Context manager to allow easy timing of events.

    This is an abstraction that needs to be implemented using a subclass
    that implements the get_current_time method.

    Usage:

        with Timer() as timer:
            do_something()
        print(timer.duration_in_s)
    """

    def __init__(self) -> None:
        # We use a stack so context managers can be nested.
        self._start_stack: list[float] = []

    def __enter__(self) -> Timer:
        self._start_stack.append(self.get_current_time())
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        end = self.get_current_time()
        start = self._start_stack.pop()
        self.duration_in_s = int(end - start)

    @abc.abstractmethod
    def get_current_time(self) -> float:
        """
        Return the time in seconds since the epoch as a floating point number.

        See https://docs.python.org/3/library/time.html#time.time
        """
        raise NotImplementedError



"""
Module for public-facing Python functions.
"""
from __future__ import annotations

from importlinter.application import use_cases

from . import configuration

configuration.configure()


def read_configuration(config_filename: str | None = None) -> dict:
    """
    Return a dictionary containing configuration from the supplied file.

    If no filename is supplied, look in the default location
    (see importlinter.cli.lint_imports).

    The dictionary has two keys:
        "session_options": dictionary of strings passed as top level configuration.
        "contracts_options": list of dictionaries, one for each contract, keyed with:
            "name": the name of the contract (str).
            "type": the type of the contract (str).
            (Any other contract-specific configuration.)

    This function is designed for use by external projects wishing to
    analyse the contracts themselves, e.g. to track the number of
    ignored imports.

    Raises:
        FileNotFoundError if no configuration file could be found.
    """
    user_options = use_cases.read_user_options(config_filename)
    return {
        "session_options": user_options.session_options,
        "contracts_options": user_options.contracts_options,
    }



from typing import Optional

from importlinter.domain.contract import Contract, ContractCheck

from . import output
from .ports.reporting import Report

# Public functions
# ----------------


def render_report(report: Report) -> None:
    """
    Output the supplied report to the console.
    """
    if report.could_not_run:
        _render_could_not_run(report)
        return

    if report.show_timings:
        output.print(f"Building graph took {report.graph_building_duration}s.")
        output.new_line()

    output.print_heading("Contracts", output.HEADING_LEVEL_TWO)
    file_count = report.module_count
    dependency_count = report.import_count
    output.print_heading(
        f"Analyzed {file_count} files, {dependency_count} dependencies.",
        output.HEADING_LEVEL_THREE,
    )

    for contract, contract_check in report.get_contracts_and_checks():
        duration = report.get_duration(contract) if report.show_timings else None
        render_contract_result_line(contract, contract_check, duration=duration)

    output.new_line()

    output.print(f"Contracts: {report.kept_count} kept, {report.broken_count} broken.")

    if report.warnings_count:
        output.new_line()
        _render_warnings(report)

    if report.broken_count:
        output.new_line()
        output.new_line()
        _render_broken_contracts_details(report)


def render_contract_result_line(
    contract: Contract, contract_check: ContractCheck, duration: Optional[int]
) -> None:
    """
    Render the one-line contract check result.

    Args:
        ...
        duration: The number of seconds the contract took to check (optional).
                  The duration will only be displayed if it is provided.
    """
    result_text = "KEPT" if contract_check.kept else "BROKEN"
    warning_text = _build_warning_text(warnings_count=len(contract_check.warnings))
    color_key = output.SUCCESS if contract_check.kept else output.ERROR
    color = output.COLORS[color_key]
    output.print(f"{contract.name} ", newline=False)
    output.print(result_text, color=color, newline=False)
    output.print(warning_text, color=output.COLORS[output.WARNING], newline=False)
    if duration is not None:
        output.print(f" [{duration}s]", newline=False)
    output.new_line()


def render_exception(exception: Exception) -> None:
    """
    Render any exception to the console.
    """
    output.print_error(str(exception))


# Private functions
# -----------------


def _render_could_not_run(report: Report) -> None:
    for contract_name, exception in report.invalid_contract_options.items():
        output.print_error(f'Contract "{contract_name}" is not configured correctly:')
        for field_name, message in exception.errors.items():
            output.indent_cursor()
            output.print_error(f"{field_name}: {message}", bold=False)


def _build_warning_text(warnings_count: int) -> str:
    if warnings_count:
        noun = "warning" if warnings_count == 1 else "warnings"
        return f" ({warnings_count} {noun})"
    else:
        return ""


def _render_warnings(report: Report) -> None:
    output.print_heading("Warnings", output.HEADING_LEVEL_TWO, style=output.WARNING)
    no_contract_outputted_yet = True

    for contract, check in report.get_contracts_and_checks():
        if check.warnings:
            if no_contract_outputted_yet:
                no_contract_outputted_yet = False
            else:
                output.new_line()
            output.print_heading(contract.name, output.HEADING_LEVEL_THREE, style=output.WARNING)
            for warning in check.warnings:
                output.print_warning(f"- {warning}")


def _render_broken_contracts_details(report: Report) -> None:
    output.print_heading("Broken contracts", output.HEADING_LEVEL_TWO, style=output.ERROR)

    for contract, check in report.get_contracts_and_checks():
        if check.kept:
            continue
        output.print_heading(contract.name, output.HEADING_LEVEL_THREE, style=output.ERROR)

        contract.render_broken_contract(check)
